// git-remote-k8s - a git remote helper to use PVCs as a remote
// Copyright (C) 2023  James Andariese
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

// TODO:
//
// * add tooling mode when it's not GIT_DIR=
// *   delete pvc
// *   check disk usage
// *   resize disk maybe?
// *   create pvc with a specific storageclass or similar
// *   transfer pvc contents
// * config mode (maybe)
// *   set the pvc storageclass instead

use default_macro::default;
use futures::{StreamExt, TryStreamExt};
use k8s::apimachinery::pkg::api::resource::Quantity;
use kube_quantity::{ParseQuantityError, ParsedQuantity};
use once_cell::sync::Lazy;
use std::collections::BTreeMap;
use std::process::ExitCode;
use std::sync::Arc;
use std::sync::atomic::{AtomicU8, Ordering};
use std::time::Duration;

use anyhow::{anyhow, bail, Result};
use thiserror::Error as ThisError;

use clap::{ArgAction, Parser as ClapParser};
use tracing::{debug, error, error_span, info, Instrument, Level};
use tracing_subscriber::FmtSubscriber;

use k8s::api::core::v1::*;
use k8s_openapi as k8s;
use kube::{api::*, config::KubeConfigOptions};

use tokio::io::{AsyncRead, AsyncReadExt, AsyncWrite, AsyncWriteExt};
use tokio::task::JoinHandle;

const KUBE_JOB_LABEL: &str = "sleeper";
const KUBE_CONTAINER_NAME: &str = KUBE_JOB_LABEL;
const SCHEME: &str = "k8s";
const PAT_SCHEME: &str = r"[a-zA-Z][a-zA-Z0-9+.-]*";
const PAT_PATH: &str = r"[0-9a-zA-Z](?:[0-9a-zA-Z.-]*[0-9a-zA-Z])?";
static REMOTE_PATTERN: Lazy<regex::Regex> = Lazy::new(|| {
    regex::Regex::new(&format!("^(?P<scheme_prefix>(?P<scheme>{PAT_SCHEME}):)?(?:/*)(?P<context>{PAT_PATH})?/(?P<namespace>{PAT_PATH})?/(?P<pvc>{PAT_PATH})?(?P<trailing>/)?")).expect("regex failed to compile")
});

#[cfg(test)]
mod test;

mod cfg;
mod errors;
mod pod_utils;
use pod_utils::is_pod_finished;
use crate::{cfg::*,errors::*,pod_utils::*};

struct AppContext {
    config: Arc<Config>,
    ensures: Vec<JoinHandle<()>>,
    rc: Arc<AtomicU8>,
}

impl Into<ExitCode> for AppContext {
    fn into(self) -> ExitCode {
        self.rc.clone().fetch_add(0, Ordering::SeqCst).into()
    }
}

impl AppContext {
    fn new() -> Result<Self> {
        Ok(Self {
            config: Arc::new(Config::parse()),
            ensures: vec![],
            rc: Arc::new(AtomicU8::new(1)),
            // ensurance: ensurance,
        })
    }
    fn cfg(&self) -> Arc<Config> {
        self.config.clone()
    }
    async fn ktx(&self, context_name: Option<String>) -> Result<kube::Client> {
        let mut kco = KubeConfigOptions::default();
        kco.context = context_name;
        Ok(kube::Client::try_from(
            kube::Config::from_kubeconfig(&kco).await?,
        )?)
    }
    async fn set_exit_code(&self, rc: u8) {
        self.rc.store(rc, Ordering::SeqCst);
    }
}

impl Drop for AppContext {
    fn drop(&mut self) {
        let handle = tokio::runtime::Handle::current();
        futures::executor::block_on(async {
            for ensure in self.ensures.drain(..) {
                if let Err(e) = handle
                    .spawn(async move {
                        let _ = ensure.await.unwrap_or_default();
                    })
                    .await
                {
                    eprintln!("failed to ensure in Context: {e}");
                }
            }
        });
    }
}

#[tokio::main]
async fn main() -> ExitCode {
    let mut ctx = {
        let ctx = AppContext::new();
        match ctx {
            Err(e) => {
                panic!("{}", e)
            },
            Ok(c) => c,
        }
    };
     
    // let cfg = ctx.cfg();

    if let Err(e) = real_main(&mut ctx).await {
        error!("{}", e);
        return ExitCode::from(127);
    }
    ctx.into()
}

async fn real_main(ctx: &mut AppContext) -> Result<()> {
    let sto = ctx.cfg().timeout;
    let sto = ctx.cfg().setup_timeout.or(sto);
    let tto = ctx.cfg().timeout;
    let tto = ctx.cfg().transfer_timeout.or(tto);

    let (pods_api, kube_worker_name) = {
        if let Some(d) = sto {
            tokio::time::timeout(d.into(), setup_pod(ctx)).await??
        } else {
            setup_pod(ctx).await?
        }
    };
    let _r = {
        if let Some(d) = tto {
            tokio::time::timeout(d.into(), do_git(ctx, pods_api, kube_worker_name)).await??
        } else {
            do_git(ctx, pods_api, kube_worker_name).await?
        }
    };
    Ok(())
    //Ok(())
}

async fn set_log_level(level: Level) {
    let subscriber = FmtSubscriber::builder()
        .with_max_level(level)
        .with_writer(std::io::stderr)
        .finish();
    tracing::subscriber::set_global_default(subscriber).expect("setting default subscriber failed");
}

async fn setup_pod(ctx: &mut AppContext) -> Result<(Api<Pod>, String)> {
    let cfg = ctx.cfg();

    set_log_level(match cfg.verbose {
        0 => Level::WARN,
        1 => Level::INFO,
        2 => Level::DEBUG,
        _ => Level::TRACE,
    })
    .await;

    debug!("{:?}", &cfg);

    // these should all be &str to ensure we don't accidentally fail to copy any
    // and make weird errors later.  instead of making these Strings here, we need
    // to to_owned() them everywhere they're referenced to make a copy to go in the
    // data structure.
    let __kube_ns_val = cfg.get_remote_namespace()?;
    let kube_ns = __kube_ns_val.as_str();
    let __kube_context_val = cfg.get_remote_context()?;
    let kube_context = __kube_context_val.as_str();
    let __kube_pvc_val = cfg.get_remote_pvc()?;
    let kube_pvc = __kube_pvc_val.as_str();
    let __kube_worker_name_val = format!("git-remote-worker-{kube_pvc}");
    let kube_worker_name = __kube_worker_name_val.as_str();
    let kube_image = cfg.image.as_str();
    let kube_shell_executable = "/bin/sh";
    let kube_shell_parameters = "-c";
    let kube_shell_sleeper_command = r#"
        set -e

        SD="$(date +%s)"
        CD=$SD
        LD=$SD
        echo $SD > /repo.timeout
        while [ $(( LD + 900 )) -ge $CD ];do
            sleep 5
            CD="$(date +%s)"
            NLD=$(cat /repo.timeout)
            if [ $NLD -ne $LD ];then
                echo "updated last run time to $NLD (from $LD)"
            fi
            LD=$NLD
        done
        echo "worker ran for $(( CD - SD )) seconds"
        "#;
    let kube_repo_mount_path = "/repo";

    let parsed_size: Result<ParsedQuantity, ParseQuantityError> =
        cfg.initial_size.clone().try_into();
    let quantity_size: Quantity = match parsed_size {
        Err(e) => {
            error!("could not parse initial PVC size: {e}");
            return Err(e.into());
        }
        Ok(s) => s.into(),
    };
    debug!("parsed size is {quantity_size:#?}");

    let kube_pod_labels = vec![
        ("com.kn8v.git-remote-k8s/repo-name", kube_pvc),
        ("com.kn8v.git-remote-k8s/job", KUBE_JOB_LABEL),
    ];

    info!("Remote Context:   {}", kube_context);
    info!("Remote Namespace: {}", kube_ns);
    info!("Remote PVC Name:  {}", kube_pvc);

    debug!("option parsing complete.  continuing to job selection.");

    if let Err(_) = std::env::var("GIT_DIR") {
        bail!("git-remote-k8s was meant to be run from git.  run with --help for more info.");
    }

    let client = ctx.ktx(Some(kube_context.to_owned())).await?;

    let pvcs_api = kube::Api::<PersistentVolumeClaim>::namespaced(client.clone(), &kube_ns);
    let pods_api = kube::Api::<Pod>::namespaced(client.clone(), &kube_ns);

    let existing_pvc = pvcs_api.get_opt(kube_pvc).await?;
    if let None = existing_pvc {
        info!("pvc doesn't exist yet.  creating now.");
        let repo_pvc = PersistentVolumeClaim {
            metadata: default!(ObjectMeta{
                name: Some(kube_pvc.to_owned()),
                namespace: Some(kube_ns.to_owned()),
            }),
            spec: Some(default!(PersistentVolumeClaimSpec {
                access_modes: Some(vec!["ReadWriteOnce".to_owned()]),
                resources: Some(ResourceRequirements {
                    claims: None,
                    limits: None,
                    requests: Some(BTreeMap::from([("storage".to_owned(), quantity_size)])),
                }),
                storage_class_name: cfg.storageclass.clone(),
                volume_mode: Some("Filesystem".to_owned()),
            })),
            status: None,
        };
        let pp = PostParams::default();
        let created_pvc = pvcs_api.create(&pp, &repo_pvc).await?;
        debug!("created pvc: {created_pvc:#?}");
    }
    debug!("{:#?}", existing_pvc);

    // create the worker pod

    let worker_pod = default!(Pod{
        metadata: default!(ObjectMeta{
            name: Some(kube_worker_name.to_owned()),
            namespace: Some(kube_ns.to_owned()),
            labels: Some({
                let mut labels = BTreeMap::new();
                for (k, v) in kube_pod_labels.iter() {
                    let kk = k.to_owned().to_owned();
                    let vv = v.to_owned().to_owned();
                    labels.insert(kk, vv);
                }
                labels
            }),
        }),
        spec: Some(default!(PodSpec {
            restart_policy: Some("Never".to_owned()),
            containers: vec![default!(Container{
                name: KUBE_CONTAINER_NAME.to_owned(),
                command: Some(vec![
                    kube_shell_executable.to_owned(),
                    kube_shell_parameters.to_owned(),
                    kube_shell_sleeper_command.to_owned(),
                ]),
                image: Some(kube_image.to_owned()),
                working_dir: Some(kube_repo_mount_path.to_owned()),
                image_pull_policy: Some("IfNotPresent".to_owned()),
                volume_mounts: Some(vec![
                    default!(VolumeMount{
                        mount_path: kube_repo_mount_path.to_owned(),
                        name: "repo".to_owned(),
                    })
                ]),
            })],
            volumes: Some(vec![
                default!(Volume{
                    name: "repo".to_owned(),
                    persistent_volume_claim: Some(
                        PersistentVolumeClaimVolumeSource {
                            claim_name: kube_pvc.to_owned(),
                            read_only: Some(false),
                        }
                    ),
                }),
            ]),
        })),
    });

    let mut lp = ListParams::default();
    let mut ls: String = String::with_capacity(kube_pod_labels.len() * 100);
    for (k, v) in kube_pod_labels {
        if ls.len() > 0 {
            ls.push(',');
        }
        ls.push_str(format!("{}={}", k.to_owned(), v).as_ref());
    }
    lp = lp.labels(&ls);
    debug!("list params: {lp:#?}");

    let worker_pods = pods_api.list(&lp).await?;

    // 1: if there is >1 pod, bail
    // 2: if there is 1 pod and that pod is running or pending, use it
    // 3: if there is 1 pod and the pod has ended, delete it and create a new one
    // 4: if there are 0 pods, start one

    let worker_pods = worker_pods.items;
    let mut opod: Option<Pod> = None;

    // 1
    if worker_pods.len() > 1 {
        error!("GIT-REMOTE CLUSTER IS IN AN INCONSISTENT STATE");
        error!("Your cluster has multiple pods running which are uniquely used for this repo.");
        let mut i = 0;
        for pod in worker_pods.iter() {
            i += 1;
            let pn = pod.metadata.name.as_ref();
            error!("pod {i}: {:?}", pn);
        }
        error!("Cannot continue while these pods are all running.");
        bail!(ApplicationError::RemoteClusterInconsistent);
    } else if worker_pods.len() == 1 {
        // 3
        let p = worker_pods
            .iter()
            .next()
            .expect("failed to take an item from an iter which is known to have enough items")
            .to_owned();
        if is_pod_finished(&pods_api, &p).await? {
            info!("existing worker is finished.  deleting.");
            let pn = p
                .metadata
                .name
                .expect("trying to delete an ended pod which has no name");
            pods_api
                .delete(&pn, &DeleteParams::default())
                .await
                .expect("failed to delete existing ended pod");
            let mut sleeptime = 0.2;
            loop {
                tokio::time::sleep(Duration::from_secs_f64(sleeptime)).await;
                if pods_api.get_opt(&pn).await?.is_none() {
                    break;
                }
                if sleeptime >= 5.0 {
                    info!("still waiting for pod to be deleted.  Ctrl-C to cancel.")
                }
                sleeptime *= 1.7;
                if sleeptime >= 10.0 {
                    sleeptime = 10.0;
                }
            }
            info!("deletion complete.  recreating worker.");
            let created_pod = pods_api.create(&PostParams::default(), &worker_pod).await?;
            opod = Some(created_pod);
        } else {
            // 2
            opod = Some(p);
        }
    } else if worker_pods.len() == 0 {
        // 4
        let created_pod = pods_api.create(&PostParams::default(), &worker_pod).await?;
        opod = Some(created_pod);
    }

    let pod: Pod;
    if let Some(p) = opod {
        pod = p;
    } else {
        error!(
            "could not find or start a worker pod (pod name: {})",
            kube_worker_name
        );
        bail!("could not find or start a worker pod");
    }

    info!("waiting for worker pod to come online");
    wait_for_pod_running(&pods_api, &pod).await?;
    info!("work pod is online.  continuing.");
    Ok((pods_api, kube_worker_name.to_owned()))
}

async fn do_git(ctx: &AppContext, pods_api: Api<Pod>, kube_worker_name: String) -> Result<()> {
    let mut gitcommand = "date +%s > /repo.timeout; 1>&2 echo welcome from the git-remote-k8s worker pod ; [ -f HEAD ] || git init --bare 1>&2".to_owned();
    let mut ttyout = tokio::io::stdout();
    let mut ttyin = tokio::io::stdin();

    let connect_cmd = negotiate_git_protocol(&mut ttyout, &mut ttyin)
        .await?
        .ok_or(anyhow!(
            "no connect command specified and we don't know how to do anything else"
        ))?;

    gitcommand.push_str(&format!(
        ";echo;{connect_cmd} .;RC=$?;1>&2 echo remote worker exiting;exit $RC"
    ));
    let ap = AttachParams::default()
        .stdin(true)
        .stdout(true)
        .stderr(true)
        .container(KUBE_CONTAINER_NAME.to_owned());
    let mut stuff = pods_api
        .exec(&kube_worker_name, vec!["sh", "-c", &gitcommand], &ap)
        .await?;
    let mut podout = stuff
        .stdout()
        .ok_or(ApplicationError::PodCouldNotOpenStdout)?;
    let mut podin = stuff
        .stdin()
        .ok_or(ApplicationError::PodCouldNotOpenStdin)?;
    // pod stderr is handled specially
    let poderr = stuff
        .stderr()
        .ok_or(ApplicationError::PodCouldNotOpenStderr)?;
    let mut poderr = tokio_util::io::ReaderStream::new(poderr);

    let barrier = Arc::new(tokio::sync::Barrier::new(4));

    let xbarrier: Arc<tokio::sync::Barrier> = barrier.clone();
    let _jhe = tokio::spawn(
        async move {
            debug!("entering");
            while let Some(l) = poderr.next().await {
                if let Err(e) = l {
                    error!("error reading from pod stderr {e}");
                    break;
                } else if let Ok(l) = l {
                    let l = String::from_utf8_lossy(l.as_ref());
                    let l = l.trim();
                    info!("from pod: {l}");
                }
            }
            debug!("waiting for group to exit");
            xbarrier.wait().await;
            debug!("exiting");
        }
        .instrument(error_span!("pod->tty", "test" = "fred")),
    );

    let xbarrier: Arc<tokio::sync::Barrier> = barrier.clone();
    let _jhi = tokio::spawn(
        async move {
            debug!("entering");
            tokio::io::copy(&mut ttyin, &mut podin)
                .await
                .expect("error copying tty input to pod input");
            podin.flush().await.expect("final flush to pod failed");
            debug!("waiting for group to exit");
            xbarrier.wait().await;
            debug!("exiting");
        }
        .instrument(error_span!("git->pod")),
    );
    let xbarrier: Arc<tokio::sync::Barrier> = barrier.clone();
    let _jho = tokio::spawn(
        async move {
            debug!("entering");
            tokio::io::copy(&mut podout, &mut ttyout)
                .await
                .expect("error copying pod output to tty output");
            ttyout.flush().await.expect("final flush to git failed");
            debug!("waiting for group to exit");
            xbarrier.wait().await;
            debug!("exiting");
        }
        .instrument(error_span!("git<-pod")),
    );

    let status = stuff
        .take_status()
        .expect("failed to take status")
        .await
        .ok_or(anyhow!("could not take status of remote git worker"))?;
    // this is an exit code which is always nonzero.
    // we'll _default_ to 1 instead of 0 because we only return _anything_ other than 0
    // when NonZeroExitCode is also given as the exit reason.
    debug!("exit code of job: {status:#?}");
    let exitcode = (|| -> Option<u8> {
        let exitcode = status
            .details
            .as_ref()?
            .causes
            .as_ref()?
            .first()?
            .message
            .to_owned()?;
        if let Ok(rc) = exitcode.parse::<u8>() {
            return Some(rc);
        }
        return Some(1);
    })()
    .unwrap_or(1);
    debug!("exit status code of remote job discovered was {exitcode:?}");
    // finally, we'll set the exit code of the entire application
    // to the exit code of the pod, if possible.  if we know it's
    // non-zero and can't figure out what it was, it will be 1.
    // if we know it's zero (because it's not non-zero), then we
    // return 0.  if we _can_ figure it out though, we use the
    // integer exit code.
    if status.reason == Some("NonZeroExitCode".to_owned()) {
        info!("exit status of remote job: {}", exitcode);
        ctx.set_exit_code(exitcode).await;
    } else {
        ctx.set_exit_code(0).await;
    }
    debug!("waiting for group to exit");
    barrier.wait().await;
    let _ = stuff.join().await?;

    Ok(())
}

async fn get_tokio_line_by_bytes(podout: &mut (impl AsyncRead + Unpin)) -> Result<String> {
    let mut r = String::with_capacity(512);

    loop {
        let c = podout.read_u8().await?;
        if c == b'\n' {
            return Ok(r);
        }
        r.push(c.into())
    }
}

#[derive(ThisError, Debug)]
enum ProtocolError {
    #[error("no command given via git protocol")]
    NoCommandGiven,
    #[error("no service given to connect command")]
    NoServiceGiven,
    #[error("unknown command specified")]
    UnknownCommand(String),
}

async fn negotiate_git_protocol(
    ttytx: &mut (impl AsyncWrite + Unpin),
    ttyrx: &mut (impl AsyncRead + Unpin),
) -> Result<Option<String>> {
    loop {
        let cmd = get_tokio_line_by_bytes(ttyrx).await?;
        let mut argv = cmd.split_whitespace();
        let cmd = argv.next().ok_or(ProtocolError::NoCommandGiven)?;

        match cmd {
            "capabilities" => {
                ttytx.write_all(b"connect\n\n").await?;
            }
            "connect" => {
                let service = argv.next().ok_or(ProtocolError::NoServiceGiven)?;
                return Ok(Some(service.to_owned()));
            }
            unknown => {
                return Err(anyhow!(ProtocolError::UnknownCommand(unknown.to_owned())));
            }
        }
    }
}
